﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.IO.MemoryMappedFiles;
using System.Runtime.InteropServices;
using System.Text;

namespace DownloadDAC
{
    public static class DacHelper
    {
        #region imports
        private const int MAX_PATH = 260;

        private const string DBGHELP = "dbghelp";

        [DllImport(DBGHELP, SetLastError = true, CharSet = CharSet.Unicode)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool SymInitialize(IntPtr hProcess,
                                                [MarshalAs(UnmanagedType.LPWStr)] string UserSearchPath,
                                                [MarshalAs(UnmanagedType.Bool)] bool fInvadeProcess);

        [DllImport(DBGHELP, SetLastError = true, CharSet = CharSet.Unicode)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool SymFindFileInPath(IntPtr hprocess,
                                                    [MarshalAs(UnmanagedType.LPWStr)] string SearchPath,
                                                    [MarshalAs(UnmanagedType.LPWStr)] string FileName,
                                                    IntPtr id,
                                                    uint two,
                                                    uint three,
                                                    uint flags,
                                                    [MarshalAs(UnmanagedType.LPWStr)] StringBuilder FoundFile,
                                                    IntPtr callback,
                                                    IntPtr context);

        [DllImport(DBGHELP)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool MiniDumpReadDumpStream(IntPtr pMiniDump,
                                                          MINIDUMP_STREAM_TYPE streamType,
                                                          out IntPtr dir,
                                                          out IntPtr streamPointer,
                                                          out uint cbStreamSize);

        private enum MINIDUMP_STREAM_TYPE
        {
            UnusedStream = 0,
            ReservedStream0 = 1,
            ReservedStream1 = 2,
            ThreadListStream = 3,
            ModuleListStream = 4,
            MemoryListStream = 5,
            ExceptionStream = 6,
            SystemInfoStream = 7,
            ThreadExListStream = 8,
            Memory64ListStream = 9,
            CommentStreamA = 10,
            CommentStreamW = 11,
            HandleDataStream = 12,
            FunctionTableStream = 13,
            UnloadedModuleListStream = 14,
            MiscInfoStream = 15,
            MemoryInfoListStream = 16,
            ThreadInfoListStream = 17,
            LastReservedStream = 0xffff,
        }

        private enum ProcessorArchitecture : ushort
        {
            PROCESSOR_ARCHITECTURE_INTEL = 0,
            PROCESSOR_ARCHITECTURE_MIPS = 1,
            PROCESSOR_ARCHITECTURE_ALPHA = 2,
            PROCESSOR_ARCHITECTURE_PPC = 3,
            PROCESSOR_ARCHITECTURE_SHX = 4,
            PROCESSOR_ARCHITECTURE_ARM = 5,
            PROCESSOR_ARCHITECTURE_IA64 = 6,
            PROCESSOR_ARCHITECTURE_ALPHA64 = 7,
            PROCESSOR_ARCHITECTURE_MSIL = 8,
            PROCESSOR_ARCHITECTURE_AMD64 = 9,
            PROCESSOR_ARCHITECTURE_IA32_ON_WIN64 = 10,
            //
            PROCESSOR_ARCHITECTURE_UNKOWN = ushort.MaxValue
        }

        [StructLayout(LayoutKind.Sequential)]
        private unsafe struct MINIDUMP_SYSTEM_INFO
        {
            internal ProcessorArchitecture ProcessorArchitecture;
            internal ushort ProcessorLevel;
            internal ushort ProcessorRevision;
            internal byte NumberOfProcessors;
            internal byte ProductType;
            internal uint MajorVersion;
            internal uint MinorVersion;
            internal uint BuildNumber;
            internal System.PlatformID PlatformId;
            internal uint CSDVersionRva;
            internal uint Reserved1;
            internal fixed uint VendorId[3];
            internal uint VersionInformation;
            internal uint FeatureInformation;
            internal uint AMDExtendedCpuFeatures;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct VS_FIXEDFILEINFO
        {
            internal uint dwSignature;
            internal uint dwStrucVersion;
            internal uint dwFileVersionMS;
            internal uint dwFileVersionLS;
            internal uint dwProductVersionMS;
            internal uint dwProductVersionLS;
            internal uint dwFileFlagsMask;
            internal uint dwFileFlags;
            internal uint dwFileOS;
            internal uint dwFileType;
            internal uint dwFileSubtype;
            internal uint dwFileDateMS;
            internal uint dwFileDateLS;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct MINIDUMP_LOCATION_DESCRIPTOR
        {
            internal uint DataSize;
            internal uint Rva;
        }

        [StructLayout(LayoutKind.Sequential, Pack = 4)]
        private struct MINIDUMP_MODULE
        {
            internal ulong baseofimage;
            internal uint SizeOfImage;
            internal uint CheckSum;
            internal uint TimeDateStamp;
            internal uint ModuleNameRva;
            internal VS_FIXEDFILEINFO VersionInfo;
            internal MINIDUMP_LOCATION_DESCRIPTOR CvRecord;
            internal MINIDUMP_LOCATION_DESCRIPTOR MiscRecord;
            internal ulong Reserved0;
            internal ulong Reserved1;
        }
        #endregion

        private class ModuleInfo
        {
            internal Version version;
            internal uint timestamp;
            internal uint size;
        }

        public static void DownloadDac(string dumpfile, string outputDirectory)
        {
            DownloadDacInternal(dumpfile, outputDirectory, "clr.dll");
        }

        private static int DownloadDacInternal(string dumpfile, string outputDirectory, string clrModulename)
        {
            var errorCode = -1;

            string archName;
            ModuleInfo moduleInfo;

            using (var mmf = MemoryMappedFile.CreateFromFile(dumpfile, FileMode.Open))
            {
                using (var accessor = mmf.CreateViewAccessor(0, 0))
                {
                    var ptr = accessor.SafeMemoryMappedViewHandle.DangerousGetHandle();

                    int x = Marshal.SizeOf(typeof(MINIDUMP_SYSTEM_INFO));

                    var procArch = GetProcessArchitecture(ptr);

                    archName = GetArchitectureName(procArch); 
                    
                    moduleInfo = GetModuleInfo(ptr, clrModulename);
                }
            }

            if ((archName != null) && (moduleInfo != null))
            {
                var dacModulename = string.Format("mscordacwks_{0}_{0}_{1}.dll", archName, moduleInfo.version);

                errorCode = DownloadModule(outputDirectory, dacModulename, moduleInfo.timestamp, moduleInfo.size);
            }

            return errorCode;
        }

        private unsafe static ProcessorArchitecture GetProcessArchitecture(IntPtr ptrDump)
        {
            var processArchitecture = ProcessorArchitecture.PROCESSOR_ARCHITECTURE_UNKOWN;

            IntPtr ptrDir, ptrStream;
            uint cbStreamSize;

            var success = MiniDumpReadDumpStream(ptrDump, MINIDUMP_STREAM_TYPE.SystemInfoStream, out ptrDir, out ptrStream, out cbStreamSize);

            if (success)
            {
                var info = (MINIDUMP_SYSTEM_INFO)Marshal.PtrToStructure(ptrStream, typeof(MINIDUMP_SYSTEM_INFO));

                processArchitecture = info.ProcessorArchitecture;
            }

            return processArchitecture;
        }

        private unsafe static ModuleInfo GetModuleInfo(IntPtr ptrDump, string moduleName)
        {
            ModuleInfo moduleInfo = null;

            IntPtr ptrDir, ptrStream;
            uint cbStreamSize;

            var success = MiniDumpReadDumpStream(ptrDump, MINIDUMP_STREAM_TYPE.ModuleListStream, out ptrDir, out ptrStream, out cbStreamSize);

            if (success)
            {
                var moduleCount = Marshal.ReadInt32(ptrStream);

                for (var i = 0; i < moduleCount; i++)
                {
                    var currentModulePtr = new IntPtr(((byte*)ptrStream) + sizeof(int) + sizeof(MINIDUMP_MODULE) * i);
                    var currentModule = (MINIDUMP_MODULE)Marshal.PtrToStructure(currentModulePtr, typeof(MINIDUMP_MODULE));

                    var currentModuleNamePtr = new IntPtr(((byte*)ptrDump) + currentModule.ModuleNameRva);
                    var currentModuleNameLength = Marshal.ReadInt32(currentModuleNamePtr);
                    var currentModuleName = Marshal.PtrToStringAuto(new IntPtr(((byte*)currentModuleNamePtr) + sizeof(int)));

                    if (Path.GetFileName(currentModuleName) == moduleName)
                    {
                        var major = (int)(currentModule.VersionInfo.dwFileVersionMS >> 16);
                        var minor = (int)(currentModule.VersionInfo.dwFileVersionMS & 0xFFFF);
                        var build = (int)(currentModule.VersionInfo.dwFileVersionLS >> 16);
                        var revision = (int)(currentModule.VersionInfo.dwFileVersionLS & 0xFFFF);

                        moduleInfo = new ModuleInfo
                        {
                            version = new Version(major, minor, build, revision),
                            timestamp = currentModule.TimeDateStamp,
                            size = currentModule.SizeOfImage
                        };

                        break;
                    }
                }
            }

            return moduleInfo;
        }

        private unsafe static int DownloadModule(string outputDirectory, string moduleName, uint moduleTimestamp, uint moduleSize)
        {
            var errorCode = 0;

            const uint SSRVOPT_DWORD = 2; 
            
            using (var currentProcess = Process.GetCurrentProcess())
            {
                var cache = @"SRV*" + outputDirectory + "*http://msdl.microsoft.com/download/symbols";

                if (SymInitialize(currentProcess.Handle, cache, false))
                {
                    var sb = new StringBuilder(MAX_PATH);
                    
                    if (!SymFindFileInPath(currentProcess.Handle, null, moduleName, new IntPtr(moduleTimestamp), moduleSize, 0, SSRVOPT_DWORD, sb, IntPtr.Zero, IntPtr.Zero))
                    {
                        errorCode = Marshal.GetLastWin32Error();
                    }
                }
            }

            return errorCode;
        }

        private static string GetArchitectureName(ProcessorArchitecture arch)
        {
            string name;
            
            switch (arch)
            {
                case ProcessorArchitecture.PROCESSOR_ARCHITECTURE_INTEL:
                    name = "x86";
                    break;

                case ProcessorArchitecture.PROCESSOR_ARCHITECTURE_AMD64:
                    name = "amd64";
                    break;

                default:
                    name = null;
                    break;
            }

            return name;
        }
    }
}
